@@ -5,7 +5,7 @@
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>
 ;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
 ;; URL: https://github.com/protesilaos/pulsar
-;; Version: 1.1.0
+;; Version: 1.2.0
 ;; Package-Requires: ((emacs "28.1"))
 ;; Keywords: convenience, pulse, highlight
 
@@ -26,46 +26,13 @@
 
 ;;; Commentary:
 ;;
-;; This is a small package that temporarily highlights the current line
-;; after a given function is invoked.  The affected functions are
-;; defined in the user option `pulsar-pulse-functions' and the effect
-;; takes place when either `pulsar-mode' (buffer-local) or
-;; `pulsar-global-mode' is enabled.
+;; This is a small package that temporarily highlights the current
+;; line after a given function is invoked.  Consult the official
+;; manual for the technicalities.
 ;;
-;; By default, Pulsar does not try behave the same way for a
-;; function's aliases.  If those are not added explicitly to the
-;; `pulsar-pulse-functions', they will not have a pulse effect.
-;; However, the user option `pulsar-resolve-pulse-function-aliases'
-;; can be set to a non-nil value to change this behaviour, meaning
-;; that Pulsar will cover a function's aliases even if those are not
-;; explicitly added to the `pulsar-pulse-functions'.
-;;
-;; The overall duration of the highlight is determined by a combination
-;; of `pulsar-delay' and `pulsar-iterations'.  The latter determines the
-;; number of blinks in a pulse, while the former sets their delay in
-;; seconds before they fade out.  The applicable face is specified in
-;; `pulsar-face'.
-;;
-;; To disable the pulse but keep the temporary highlight, set the user
-;; option `pulsar-pulse' to nil.  The current line will remain
-;; highlighted until another command is invoked.
-;;
-;; To highlight the current line on demand, use the `pulsar-pulse-line'
-;; command.  When `pulsar-pulse' is non-nil (the default), its highlight
-;; will pulse before fading away.  Whereas the `pulsar-highlight-line'
-;; command never pulses the line: the highlight stays in place as if
-;; `pulsar-pulse' is nil.
-;;
-;; To help users differentiate between the pulse and highlight effects,
-;; the user option `pulsar-highlight-face' controls the presentation of
-;; the `pulsar-highlight-line' command.  By default, this variable is
-;; the same as `pulsar-face'.
-;;
-;; Pulsar depends on the built-in `pulse.el' library.
-;;
-;; Why the name "pulsar"?  It sounds like "pulse" and is a recognisable
-;; word.  Though if you need a backronym, consider "Pulsar Unquestionably
-;; Luminates, Strictly Absent the Radiation".
+;; Why the name "pulsar"?  It sounds like "pulse" and is a
+;; recognisable word.  Though if you need a backronym, consider
+;; "Pulsar Unquestionably Luminates, Strictly Absent the Radiation".
 
 ;;; Code:
 
@@ -78,6 +45,30 @@ Extension of `pulse.el'."
 
 ;;;; User options
 
+(defcustom pulsar-pulse t
+  "When non-nil enable pulsing.
+Otherwise the highlight stays on the current line until another
+command is invoked."
+  :type 'boolean
+  :package-version '(pulsar . "0.2.0")
+  :group 'pulsar)
+
+(defcustom pulsar-delay 0.05
+  "Delay between increments of a pulse.
+Together with `pulsar-iterations' control the overall duration of
+a pulse.  Only applies when `pulsar-pulse' is non-nil."
+  :type 'number
+  :package-version '(pulsar . "0.1.0")
+  :group 'pulsar)
+
+(defcustom pulsar-iterations pulse-iterations
+  "Number of iterations in a pulse highlight.
+Together with `pulsar-delay' control the overall duration of a
+pulse.  Only applies when `pulsar-pulse' is non-nil."
+  :type 'number
+  :package-version '(pulsar . "0.1.0")
+  :group 'pulsar)
+
 (defcustom pulsar-pulse-functions
   '(backward-page
     bookmark-jump
@@ -139,16 +130,86 @@ This only takes effect when `pulsar-mode' (buffer-local) or
   :package-version '(pulsar . "1.1.0")
   :group 'pulsar)
 
+(defvar pulsar-pulse-region-common-functions
+  '(yank
+    kill-region
+    kill-line
+    kill-ring-save
+    append-next-kill
+    kill-whole-line
+    kill-visual-line
+    kill-word backward-kill-word
+    kill-sentence backward-kill-sentence
+    kill-paragraph backward-kill-paragraph
+    kill-sexp backward-kill-sexp
+    kill-rectangle
+    yank-rectangle
+    open-rectangle
+    undo)
+  "Common functions that can be used for `pulsar-pulse-region-functions'.")
+
+(defcustom pulsar-pulse-region-functions nil
+  "Functions that highlight the affected region after invocation.
+When the value is nil, no pulsing is in effect.  Otherwise, the value is
+a list of functions that operate on the region.  It can be, for example,
+`pulsar-pulse-region-common-functions'.
+
+This only takes effect when `pulsar-mode' (buffer-local) or
+`pulsar-global-mode' is enabled."
+  :type '(choice
+          (const :tag "Do not pulse the region" nil)
+          (repeat function))
+  :package-version '(pulsar . "1.2.0")
+  :group 'pulsar)
+
 (defcustom pulsar-resolve-pulse-function-aliases t
-  "When non-nil, resolve function aliases in `pulsar-pulse-functions'.
-This allows pulsar to respect, e.g., `tab-new' \"parent,\"
-`tab-bar-new-tab', and vice-versa, enabling Pulsar to respect
-`tab-bar-new-tab' alias `tab-new'."
+  "When non-nil, resolve function aliases in pulsar function lists.
+The affected lists are `pulsar-pulse-functions' and
+`pulsar-pulse-region-functions'.  This allows pulsar to respect,
+e.g., `tab-new' \"parent,\" `tab-bar-new-tab', and vice-versa,
+enabling Pulsar to respect `tab-bar-new-tab' alias `tab-new'."
   :type 'boolean
   :package-version '(pulsar . "1.1.0")
   :group 'pulsar)
 
-(make-obsolete 'pulsar-pulse-on-window-change nil "0.5.0")
+(defcustom pulsar-inhibit-hidden-buffers t
+  "When non-nil, `pulsar-mode' will not be enabled in hidden buffers.
+Hidden buffers are those whose name starts with a space character.  They
+are not meant to be touched by the user, so pulsing in them is not
+necessary.  This option is provided in case there is some scenario where
+pulsing makes sense."
+  :type 'boolean
+  :package-version '(pulsar . "1.2.0")
+  :group 'pulsar)
+
+(defcustom pulsar-pulse-on-window-change nil
+  "When non-nil, enable pulsing on every window change.
+This covers all commands or functions that affect the current
+window.  Users who prefer to trigger a pulse only after select
+functions (e.g. only after `other-window') are advised to set
+this variable to nil and update the `pulsar-pulse-functions'
+accordingly."
+  :type 'boolean
+  :package-version '(pulsar . "1.2.0")
+  :group 'pulsar)
+
+(defconst pulsar--face-choice-widget
+  '(radio :tag "Select another face"
+          (face :tag "Generic pulse.el face" pulsar-generic)
+          (face :tag "Red style" pulsar-red)
+          (face :tag "Green style" pulsar-green)
+          (face :tag "Yellow style" pulsar-yellow)
+          (face :tag "Blue style" pulsar-blue)
+          (face :tag "Magenta style" pulsar-magenta)
+          (face :tag "Cyan style" pulsar-cyan)
+          (face :tag "Other face (must have a background)"))
+  "Widget for `defcustom' type to select a face.")
+
+(defconst pulsar--face-with-default-and-choice-widget
+  `(choice :tag "Choose a face"
+           (variable :tag "Use the `pulsar-face'" pulsar-face)
+           ,pulsar--face-choice-widget)
+  "Like `pulsar--face-choice-widget' plus the `pulsar-face' option.")
 
 (defcustom pulsar-face 'pulsar-generic
   "Face of the regular pulse line effect (`pulsar-pulse-line').
@@ -157,53 +218,32 @@ from the underlying pulse library.  Users can select one among
 `pulsar-red', `pulsar-green', `pulsar-yellow', `pulsar-blue',
 `pulsar-magenta', `pulsar-cyan', or any other face that has a
 background attribute."
-  :type '(radio (face :tag "Generic pulse.el face" pulsar-generic)
-                (face :tag "Red style" pulsar-red)
-                (face :tag "Green style" pulsar-green)
-                (face :tag "Yellow style" pulsar-yellow)
-                (face :tag "Blue style" pulsar-blue)
-                (face :tag "Magenta style" pulsar-magenta)
-                (face :tag "Cyan style" pulsar-cyan)
-                (face :tag "Other face (must have a background)"))
+  :type pulsar--face-choice-widget
   :package-version '(pulsar . "0.2.0")
   :group 'pulsar)
 
 (defcustom pulsar-highlight-face 'pulsar-face
   "Face used in `pulsar-highlight-line'."
-  :type '(choice (variable pulsar-face)
-                 (radio (face :tag "Generic pulse.el face" pulsar-generic)
-                        (face :tag "Red style" pulsar-red)
-                        (face :tag "Green style" pulsar-green)
-                        (face :tag "Yellow style" pulsar-yellow)
-                        (face :tag "Blue style" pulsar-blue)
-                        (face :tag "Magenta style" pulsar-magenta)
-                        (face :tag "Cyan style" pulsar-cyan)
-                        (face :tag "Other face (must have a background)")))
+  :type pulsar--face-with-default-and-choice-widget
   :package-version '(pulsar . "0.3.0")
   :group 'pulsar)
 
-(defcustom pulsar-pulse t
-  "When non-nil enable pulsing.
-Otherwise the highlight stays on the current line until another
-command is invoked."
-  :type 'boolean
-  :package-version '(pulsar . "0.2.0")
+(defcustom pulsar-region-face pulsar-face
+  "Face to pulse a region that has not changed."
+  :type pulsar--face-with-default-and-choice-widget
+  :package-version '(pulsar . "1.2.0")
   :group 'pulsar)
 
-(defcustom pulsar-delay 0.05
-  "Delay between increments of a pulse.
-Together with `pulsar-iterations' control the overall duration of
-a pulse.  Only applies when `pulsar-pulse' is non-nil."
-  :type 'number
-  :package-version '(pulsar . "0.1.0")
+(defcustom pulsar-region-change-face pulsar-face
+  "Face to pulse a region that has changed (added or removed)."
+  :type pulsar--face-with-default-and-choice-widget
+  :package-version '(pulsar . "1.2.0")
   :group 'pulsar)
 
-(defcustom pulsar-iterations pulse-iterations
-  "Number of iterations in a pulse highlight.
-Together with `pulsar-delay' control the overall duration of a
-pulse.  Only applies when `pulsar-pulse' is non-nil."
-  :type 'number
-  :package-version '(pulsar . "0.1.0")
+(defcustom pulsar-window-change-face pulsar-face
+  "Face to indicate the current position on window changes."
+  :type pulsar--face-with-default-and-choice-widget
+  :package-version '(pulsar . "1.2.0")
   :group 'pulsar)
 
 ;;;; Faces
@@ -284,9 +324,9 @@ pulse.  Only applies when `pulsar-pulse' is non-nil."
   (save-excursion
     (goto-char (line-beginning-position))
     (and (not (bobp))
-	     (or (beginning-of-line 1) t)
-	     (save-match-data
-	       (looking-at "^[\s\t]+")))))
+         (or (beginning-of-line 1) t)
+         (save-match-data
+           (looking-at "^[\s\t]+")))))
 
 (defun pulsar--buffer-end-p ()
   "Return non-nil if point is at the end of the buffer."
@@ -325,6 +365,8 @@ With optional FACE, use it instead of `pulsar-face'.
 
 With optional START and END, highlight the region in-between
 instead of the current line."
+  (when (and (numberp start) (numberp end) (= start end)) ; pulse the whole line if start=end
+    (setq start nil end nil))
   (let* ((pulse-flag (if no-pulse nil pulsar-pulse))
          (pulse-delay pulsar-delay)
          (pulse-iterations pulsar-iterations)
@@ -350,27 +392,28 @@ pulse effect."
 (defun pulsar-pulse-region ()
   "Temporarily highlight the active region if any."
   (interactive)
-  (when (region-active-p)
-    (let ((beg (region-beginning))
-          (end (region-end)))
-      ;; FIXME 2024-08-29: Finding the lines and columns therein
-      ;; does not work because consecutive pulses cancel each
-      ;; other out, leaving only the last one active.
-      ;;
-      ;; (let* ((columns (rectangle--pos-cols beg end))
-      ;;        (begcol (car columns))
-      ;;        (endcol (cdr columns)))
-      ;;    (lines (list
-      ;;            (line-number-at-pos beg)
-      ;;            (line-number-at-pos end))))
-      ;; (dolist (line lines)
-      ;;   (save-excursion
-      ;;     (goto-char (point-min))
-      ;;     (forward-line (1- line))
-      ;;     (setq beg (progn (move-to-column begcol) (point))
-      ;;           end (progn (move-to-column endcol) (point))))
-      ;;   (pulsar--pulse nil nil beg end)))
-      (pulsar--pulse nil nil beg end))))
+  (if (region-active-p)
+      (let ((beg (region-beginning))
+            (end (region-end)))
+        ;; FIXME 2024-08-29: Finding the lines and columns therein
+        ;; does not work because consecutive pulses cancel each
+        ;; other out, leaving only the last one active.
+        ;;
+        ;; (let* ((columns (rectangle--pos-cols beg end))
+        ;;        (begcol (car columns))
+        ;;        (endcol (cdr columns)))
+        ;;    (lines (list
+        ;;            (line-number-at-pos beg)
+        ;;            (line-number-at-pos end))))
+        ;; (dolist (line lines)
+        ;;   (save-excursion
+        ;;     (goto-char (point-min))
+        ;;     (forward-line (1- line))
+        ;;     (setq beg (progn (move-to-column begcol) (point))
+        ;;           end (progn (move-to-column endcol) (point))))
+        ;;   (pulsar--pulse nil nil beg end)))
+        (pulsar--pulse nil pulsar-region-face beg end))
+    (pulsar--pulse nil pulsar-region-face)))
 
 ;;;###autoload
 (defun pulsar-highlight-line ()
@@ -460,7 +503,7 @@ For lines, do the same as `pulsar-highlight-line'."
   (cond
    ((bound-and-true-p rectangle-mark-mode)
     (pulsar--highlight-rectangle))
-   ((use-region-p)
+   ((region-active-p)
     (pulsar--pulse :no-pulse pulsar-highlight-face (region-beginning) (region-end)))
    (t
     (pulsar--pulse :no-pulse pulsar-highlight-face))))
@@ -468,40 +511,97 @@ For lines, do the same as `pulsar-highlight-line'."
 ;;;; Mode setup
 
 (define-minor-mode pulsar-mode
-  "Set up pulsar for each function in `pulsar-pulse-functions'.
-This is a buffer-local mode.  Also check `pulsar-global-mode'."
+  "Set up pulsar for each function in pulsar functions lists.
+The effective lists are `pulsar-pulse-functions' and
+`pulsar-pulse-region-functions'.  This is a buffer-local mode.
+Also check `pulsar-global-mode'."
   :global nil
   (if pulsar-mode
       (progn
         (when pulsar-resolve-pulse-function-aliases
-          (pulsar--resolve-function-aliases))
-        (add-hook 'post-command-hook #'pulsar--post-command-pulse nil 'local))
-    (remove-hook 'post-command-hook #'pulsar--post-command-pulse 'local)))
+          (pulsar-resolve-function-aliases))
+        (add-hook 'post-command-hook #'pulsar--post-command-pulse nil 'local)
+        (add-hook 'after-change-functions #'pulsar--after-change-function nil 'local)
+        (when pulsar-pulse-on-window-change
+          (add-hook 'window-buffer-change-functions #'pulsar--pulse-on-window-change nil 'local)
+          (add-hook 'window-selection-change-functions #'pulsar--pulse-on-window-change nil 'local)))
+    (remove-hook 'post-command-hook #'pulsar--post-command-pulse 'local)
+    (remove-hook 'after-change-functions #'pulsar--after-change-function 'local)
+    (remove-hook 'window-buffer-change-functions #'pulsar--pulse-on-window-change 'local)
+    (remove-hook 'window-selection-change-functions #'pulsar--pulse-on-window-change 'local)))
 
 (defun pulsar--on ()
   "Enable `pulsar-mode'."
-  (unless (minibufferp)
+  (unless (or pulsar-mode
+              (minibufferp)
+              (and pulsar-inhibit-hidden-buffers (string-prefix-p " " (buffer-name))))
     (let (inhibit-quit)
       (pulsar-mode 1))))
 
 ;;;###autoload
 (define-globalized-minor-mode pulsar-global-mode pulsar-mode pulsar--on)
 
+(defun pulsar--pulse-on-window-change (window)
+  "Run `pulsar-pulse-line' on WINDOW change."
+  (when (and pulsar-mode
+             pulsar-pulse-on-window-change
+             (eq (frame-selected-window) window)
+             (not (minibufferp))
+             ;; Avoid double pulsing when both
+             ;; pulsar-pulse-on-window-change and
+             ;; pulsar-pulse-functions are in effect.
+             (not (memq this-command pulsar-pulse-functions))
+             (not (memq real-this-command pulsar-pulse-functions)))
+    (pulsar--pulse nil pulsar-window-change-face)))
+
+(defvar-local pulsar--pulse-region-changes nil)
+
+;; This feature is heavily inspired by Daniel Mendler's `goggles' package.
+(defun pulsar--after-change-function (beg end len)
+  "Provide `after-change-functions' hook to accumulate buffer edits.
+Changes are defined by BEG, END, LEN:
+
+- BEG and END mark the region of text.
+- LEN is zero for insertions.
+- LEN is the extent of deletions and BEG==END."
+  (when (or (memq this-command pulsar-pulse-region-functions)
+            (memq real-this-command pulsar-pulse-region-functions))
+    (when (and (zerop len) (= beg end)) ; In case of a deletion
+      (when (> beg (buffer-size))
+        (setq beg (1- beg)))
+      (setq end (1+ beg)))
+    (push (cons (copy-marker beg) (copy-marker end)) pulsar--pulse-region-changes)))
+
 (defun pulsar--post-command-pulse ()
-  "Run `pulsar-pulse-line' for `pulsar-pulse-functions'."
-  (when (and (or pulsar-mode pulsar-global-mode)
-             (or (memq this-command pulsar-pulse-functions)
-                 (memq real-this-command pulsar-pulse-functions)))
-    (pulsar-pulse-line)))
+  "Pulse current line, accumulated edits, or selected region."
+  (when pulsar-mode
+    (cond
+     ((or (memq this-command pulsar-pulse-functions)
+          (memq real-this-command pulsar-pulse-functions))
+      (pulsar-pulse-line))
+     ;; Extract the outer limits of the affected region from
+     ;; accumulated changes. NOTE: Non-contiguous regions such as
+     ;; rectangles will pulse their contiguous bounds.
+     (pulsar--pulse-region-changes
+      (let ((beg (apply #'min (mapcar #'car pulsar--pulse-region-changes)))
+            (end (apply #'max (mapcar #'cdr pulsar--pulse-region-changes))))
+        (setq pulsar--pulse-region-changes nil)
+        (pulsar--pulse nil pulsar-region-change-face beg end)))
+     ;; Pulse the selected region for commands that did not cause
+     ;; buffer changes; e.g., kill-ring-save.
+     ((or (memq this-command pulsar-pulse-region-functions)
+          (memq real-this-command pulsar-pulse-functions))
+      (pulsar-pulse-region)))))
 
 (make-obsolete 'pulsar-setup nil "0.3.0")
 
-;; Lifted from Emacs 30 to allow pulsar to remain backward compatbile
-;; with Emacs earlier than Emacs 29.1. TODO: Deprecate this at some
-;; point to prefer Emacs core.
+;; TODO 2024-11-26: Deprecate this at some point to prefer Emacs core.
 (defun pulsar--function-alias-p (func &optional _noerror)
   "Return nil if FUNC is not a function alias.
-If FUNC is a function alias, return the function alias chain."
+If FUNC is a function alias, return the function alias chain.
+
+This is a copy of `function-alias-p' for backward-compatibility and will
+be deleted from Pulsar in future versions of Emacs."
   (declare (advertised-calling-convention (func) "30.1")
            (side-effect-free error-free))
   (let ((chain nil))
@@ -511,10 +611,10 @@ If FUNC is a function alias, return the function alias chain."
       (push func chain))
     (nreverse chain)))
 
-;; This is essentially the inverse of function-alias-p for a list of
-;; function symbols.
 (defun pulsar--find-fn-aliases (fns)
-  "Return a list of aliases for the FNS symbols."
+  "Return a list of aliases for the FNS symbols.
+This is essentially the inverse of `pulsar--function-alias-p' for a list
+of function symbols."
   (let ((aliases))
     (mapatoms (lambda (sym)
                 (when (and
@@ -523,12 +623,23 @@ If FUNC is a function alias, return the function alias chain."
                   (push sym aliases))))
     aliases))
 
-(defun pulsar--resolve-function-aliases ()
-  "Amend `pulsar-pulse-functions' to respect function aliases."
-  (setq pulsar-pulse-functions
-        (seq-union pulsar-pulse-functions
-                   (seq-union (pulsar--find-fn-aliases pulsar-pulse-functions)
-                              (flatten-list (mapcar (lambda (x) (pulsar--function-alias-p x)) pulsar-pulse-functions))))))
+(defun pulsar--find-union (functions)
+  "Find the union of FUNCTIONS and their aliases."
+  (seq-union functions
+             (seq-union (pulsar--find-fn-aliases functions)
+                        (flatten-list (mapcar #'pulsar--function-alias-p functions)))))
+
+(defun pulsar-resolve-function-aliases ()
+  "Amend registered functions to respect function aliases.
+Functions are registered in `pulsar-pulse-functions' and
+`pulsar-pulse-region-functions'.  This is called automatically
+when `pulsar-resolve-pulse-function-aliases' is non-nil.
+
+You may also call this manually in your configuration after setting
+`pulsar-pulse-functions'.  In that case, you would prefer
+`pulsar-resolve-pulse-function-aliases' to be nil."
+  (setq pulsar-pulse-functions (pulsar--find-union pulsar-pulse-functions))
+  (setq pulsar-pulse-region-functions (pulsar--find-union pulsar-pulse-region-functions)))
 
 ;;;; Recentering commands
 
@@ -568,4 +679,5 @@ Use this in combination with `pulsar-recenter-top' or
     (outline-show-entry))))
 
 (provide 'pulsar)
+
 ;;; pulsar.el ends here
